/*
 * Copyright 2011 Red Hat, Inc.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include "callqueue.h"
#include "common.h"
#include <assert.h>
#include <stdlib.h>
#include <string.h>

typedef struct call call;

struct call {
  int id;
  json_object *mesg;
  call *next;
  void *misc;
  jsonrpc_free *free;
  jsonrpc_response_cb *resp;
};

struct callqueue {
  dataqueue *dq;
  call *head;
  call *tail;
  size_t len;
  json_object *batch;
};

callqueue *
callqueue_new(dataqueue *dq)
{
  callqueue *cq;

  cq = malloc(sizeof(callqueue));
  if (!cq)
    return NULL;
  memset(cq, 0, sizeof(callqueue));

  cq->dq = dq;
  return cq;
}

static void
call_free(call *cll)
{
  if (cll) {
    if (cll->free)
      cll->free(cll->misc);
    json_object_put(cll->mesg);
    free(cll);
  }
}

static void
call_free_recursive(call *cll)
{
  if (cll)
    call_free_recursive(cll->next);
  call_free(cll);
}

static call *
call_remove_notifications(call *cll, call **tail, size_t *len)
{
  call *tmp;

  if (!cll)
    return NULL;

  if (cll->resp) {
    cll->next = call_remove_notifications(cll->next, tail, len);
    if (!cll->next)
      *tail = cll;
    return cll;
  }

  tmp = cll->next;
  call_free(cll);
  (*len)--;
  return call_remove_notifications(tmp, tail, len);
}

void
callqueue_remove_notifications(callqueue *cq)
{
  if (cq) {
    cq->head = call_remove_notifications(cq->head, &cq->tail, &cq->len);
    if (!cq->head)
      cq->tail = NULL;
  }
}

bool
callqueue_push(callqueue *cq, int id, const char *method, json_object *params,
               jsonrpc_response_cb *resp, void *misc, jsonrpc_free *mfree)
{
  call *cll;
  json_object *tmp;

  if (!cq || !method
      || (params && !json_object_is_type(params, json_type_array)
                 && !json_object_is_type(params, json_type_object))
      || !(cll = malloc(sizeof(call)))) {
    if (mfree)
      (*mfree)(misc);
    return false;
  }
  memset(cll, 0, sizeof(call));

  /* Fill out our request object */
  cll->id = id;
  cll->misc = misc;
  cll->free = mfree;
  cll->resp = resp;
  cll->mesg = json_object_new_object();
  if (!cll->mesg)
    goto error;

  /* Create our object */
  tmp = json_object_new_string(JSONRPC_VERSION);
  if (!tmp)
    goto error;
  json_object_object_add(cll->mesg, "jsonrpc", tmp);
  tmp = json_object_new_string(method);
  if (!tmp)
    goto error;
  json_object_object_add(cll->mesg, "method", tmp);
  if (params)
    json_object_object_add(cll->mesg, "params", json_object_get(params));
  if (resp) {
    tmp = json_object_new_int(cll->id);
    if (!tmp)
      goto error;
    json_object_object_add(cll->mesg, "id", tmp);
  }

  /* If we have a dataqueue, push our message into it */
  if (cq->dq && !dataqueue_push(cq->dq, true,
                                json_object_to_json_string(cll->mesg)))
    goto error;

  cll->next = cq->head;
  cq->head = cll;
  cq->len++;
  if (!cq->tail)
    cq->tail = cll;

  json_object_put(params);
  return true;

error:
  call_free(cll);
  json_object_put(params);
  return false;
}

void
callqueue_move(callqueue *from, callqueue *to)
{
  assert(from && to);

  if (from->head && from->tail) {
    from->tail->next = to->head;
    to->head = from->head;
    from->head = NULL;
    from->tail = NULL;
    from->len = 0;
  }
}

const char *
callqueue_as_batch(callqueue *cq)
{
  call *cll;
  int i;

  if (!cq)
    return NULL;

  json_object_put(cq->batch);
  cq->batch = json_object_new_array();
  if (!cq->batch)
    return NULL;


  for (cll = cq->head, i = 1; cll; cll = cll->next, i++) {

    if (json_object_array_put_idx(cq->batch, cq->len - i,
                                  json_object_get(cll->mesg)) != 0) {
      json_object_put(cll->mesg);
      return NULL;
    }
  }

  return json_object_to_json_string(cq->batch);
}

bool
callqueue_respond(callqueue *cq, int id, json_object *response, bool error)
{
  call *cll, *prv;

  if (!cq)
    return false;

  for (prv = NULL, cll = cq->head; cll; prv = cll, cll = cll->next) {
    if (cll->id == id) {
      cll->resp(response, error, cll->misc);
      if (prv)
        prv->next = cll->next;
      else
        cq->head = cll->next;

      if (!cll->next)
        cq->tail = prv;

      call_free(cll);
      return true;
    }
  }

  return false;
}

void
callqueue_free(callqueue *cq)
{
  if (cq) {
    call_free_recursive(cq->head);
    json_object_put(cq->batch);
    free(cq);
  }
}

