/*
 * Copyright 2011 Red Hat, Inc.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include "dataqueue.h"
#include <stdlib.h>
#include <string.h>

typedef struct item item;
struct item {
  item *next;
  item *prev;
  char *data;
  size_t len;
};

struct dataqueue {
  item *head;
  item *tail;
  size_t len;
};

dataqueue *
dataqueue_new(void)
{
  dataqueue *queue;

  queue = malloc(sizeof(dataqueue));
  if (queue)
    memset(queue, 0, sizeof(dataqueue));
  return queue;
}

bool
dataqueue_push(dataqueue *queue, bool newl, const char *data)
{
  return dataqueue_push_len(queue, newl, data, data ? strlen(data) : 0);
}

bool
dataqueue_push_len(dataqueue *queue, bool newl, const char *data, size_t len)
{
  item *tmp;

  if (!queue || !data)
    return false;

  tmp = malloc(sizeof(item));
  if (!tmp)
    return false;
  memset(tmp, 0, sizeof(item));

  tmp->len = len;
  tmp->data = malloc(sizeof(char) * (len + 2));
  if (!tmp->data) {
    free(tmp);
    return false;
  }
  strncpy(tmp->data, data, len);
  if (newl)
    tmp->data[len++] = '\n';
  tmp->data[len] = '\0';

  tmp->next = queue->head;
  if (tmp->next)
    tmp->next->prev = tmp;
  queue->head = tmp;
  if (!queue->tail)
    queue->tail = tmp;
  queue->len++;
  return true;
}

const char *
dataqueue_peek(dataqueue *queue)
{
  return queue ? (queue->tail ? queue->tail->data : NULL) : NULL;
}

char *
dataqueue_pop(dataqueue *queue)
{
  item *tmp;
  char *data;

  if (!queue || !queue->tail)
    return NULL;

  /* Remove the last item */
  tmp = queue->tail;
  if (queue->tail->prev)
    queue->tail->prev->next = NULL;
  queue->tail = queue->tail->prev;
  if (!queue->tail)
    queue->head = NULL;

  /* Free it, but return its data */
  data = tmp->data;
  free(tmp);
  queue->len--;
  return data;
}

void
dataqueue_partial(dataqueue *queue, size_t size)
{
  if (!queue || !queue->tail)
    return;

  if (size >= queue->tail->len) {
    free(dataqueue_pop(queue));
    return;
  }

  memmove(queue->tail->data, queue->tail->data + size,
          queue->tail->len - size + 1);
}

size_t
dataqueue_len(dataqueue *queue)
{
  return queue ? queue->len : 0;
}

static void
item_free_recursive(item *i)
{
  if (i) {
    item_free_recursive(i->next);
    free(i->data);
    free(i);
  }
}

void
dataqueue_free(dataqueue *queue)
{
  item_free_recursive(queue ? queue->head : NULL);
  free(queue);
}
