/*
 * Copyright 2011 Red Hat, Inc.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include "callqueue.h"
#include "rpcobject.h"
#include "verto-jsonrpc.h"
#ifndef VERTO_EV_FLAG_IO_ERROR
#define VERTO_EV_FLAG_IO_ERROR 0
#endif

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>

#define __str(s) # s
#define _str(s) __str(s)

#define JSONRPC_VERSION "2.0"
#define JSONRPC_MESSAGE_TERMINATOR '\n'

#define JSONRPC_ERRMESG_INVALID_REQUEST  "Invalid request."
#define JSONRPC_ERRMESG_METHOD_NOT_FOUND "Method not found."
#define JSONRPC_ERRMESG_INVALID_PARAMS   "Invalid parameters."
#define JSONRPC_ERRMESG_INTERNAL_ERROR   "Internal error."
#define JSONRPC_ERRMESG_PARSE_ERROR      "Parse error."

#define JSONRPC_ERROR_INVLREQ "{\"jsonrpc\": \"" JSONRPC_VERSION "\", \"id\": null, " \
                               "\"error\": {\"code\": " _str(JSONRPC_ERRCODE_INVALID_REQUEST) ", " \
                                           "\"message\": \"" JSONRPC_ERRMESG_INVALID_REQUEST "\"}}\n"
#define JSONRPC_ERROR_INTERR "{\"jsonrpc\": \"" JSONRPC_VERSION "\", \"id\": null, " \
                              "\"error\": {\"code\": " _str(JSONRPC_ERRCODE_INTERNAL_ERROR) ", " \
                                          "\"message\": \"" JSONRPC_ERRMESG_INTERNAL_ERROR "\"}}\n"
#define JSONRPC_ERROR_PARSEERR "{\"jsonrpc\": \"" JSONRPC_VERSION "\", \"id\": null, " \
                                "\"error\": {\"code\": " _str(JSONRPC_ERRCODE_PARSE_ERROR) ", " \
                                            "\"message\": \"" JSONRPC_ERRMESG_PARSE_ERROR "\"}}\n"
#define JSONRPC_ERROR_TOOBIG "{\"jsonrpc\": \"" JSONRPC_VERSION "\", \"id\": null, " \
                              "\"error\": {\"code\": -32000, " \
                                          "\"message\": \"Request is too large.\"}}\n"

typedef struct input input;
typedef struct handler handler;
typedef struct jsonrpc jsonrpc;

/* Outoing input structures */
struct jsonrpc_batch {
  jsonrpc *rpc;
  callqueue *cq;
  jsonrpc_batch *next;
};

struct handler {
  void *misc;
  jsonrpc_free *free;
  char *method;
  jsonrpc_handler_cb *handler;
};

struct jsonrpc {
  int serial;
  jsonrpc_error_cb *errcb;
  void *misc;
  jsonrpc_free *free;
  jsonrpc_batch *batches;

  struct {
    struct {
      int fd;
      verto_ev *ev;
    } in;

    struct {
      int fd;
      verto_ev *ev;
    } out;

    struct {
      char *data;
      size_t used;
      size_t size;
    } buffer;

    struct {
      dataqueue *awaiting_write;
      callqueue *awaiting_response;
    } queues;
  } io;

  struct {
    json_tokener *tokener;
    bool msgtoolarge;
  } parser;

  struct {
    handler **data;
    size_t size;
    size_t used;
  } handlers;
};

static void
int_handler_free(handler *h)
{
  if (!h)
    return;

  if (h->free)
    h->free(h->misc);
  free(h->method);
  free(h);
}

static void
int_jsonrpc_free(jsonrpc *rpc)
{
  size_t i;

  if (!rpc)
    return;

  if (rpc->free)
    rpc->free(rpc->misc);

  if (rpc->io.in.ev != rpc->io.out.ev)
    verto_del(rpc->io.out.ev);
  free(rpc->io.buffer.data);
  dataqueue_free(rpc->io.queues.awaiting_write);
  callqueue_free(rpc->io.queues.awaiting_response);
  json_tokener_free(rpc->parser.tokener);

  for (i = 0; i < rpc->handlers.used; i++)
    int_handler_free(rpc->handlers.data[i]);
  free(rpc->handlers.data);

  free(rpc);
}

static json_object *
int_get_error(const char *mesg, json_object *id, int code, bool onlyerrobj)
{
  json_object *tmpa = NULL, *tmpb = NULL;
  static struct {
    int code;
    const char *mesg;
  } standard_errors_messages[] = {
    { JSONRPC_ERRCODE_INVALID_REQUEST, JSONRPC_ERRMESG_INVALID_REQUEST },
    { JSONRPC_ERRCODE_METHOD_NOT_FOUND, JSONRPC_ERRMESG_METHOD_NOT_FOUND },
    { JSONRPC_ERRCODE_INVALID_PARAMS, JSONRPC_ERRMESG_INVALID_PARAMS },
    { JSONRPC_ERRCODE_INTERNAL_ERROR, JSONRPC_ERRMESG_INTERNAL_ERROR },
    { JSONRPC_ERRCODE_PARSE_ERROR, JSONRPC_ERRMESG_PARSE_ERROR },
    { 0, NULL }
  };

  if (!mesg) {
    int i = 0;

    do {
      mesg = standard_errors_messages[i].mesg;
      if (standard_errors_messages[i].code == code)
        break;
    } while (standard_errors_messages[i++].mesg);

    if (!mesg)
      goto error;
  }

  /* Create error object */
  tmpb = json_object_new_object();
  tmpa = json_object_new_int(code);
  if (!tmpb || !tmpa)
    goto error;
  json_object_object_add(tmpb, "code", tmpa);
  tmpa = json_object_new_string(mesg);
  if (!tmpa)
    goto error;
  json_object_object_add(tmpb, "message", tmpa);
  if (onlyerrobj) {
    json_object_put(id);
    return tmpb;
  }

  /* Create msg_state */
  tmpa = json_object_new_object();
  if (!tmpa)
    goto error;
  json_object_object_add(tmpa, "error", tmpb);
  tmpb = json_object_new_string(JSONRPC_VERSION);
  if (!tmpb)
    goto error;
  json_object_object_add(tmpa, "jsonrpc", tmpb);
  json_object_object_add(tmpa, "id", id);

  return tmpa;

error:
  json_object_put(id);
  json_object_put(tmpa);
  json_object_put(tmpb);
  return NULL;
}

static void
int_on_writer_free(verto_ctx *ctx, verto_ev *ev)
{
  jsonrpc *rpc;

  rpc = verto_get_private(ev);
  if (rpc)
    rpc->io.out.ev = NULL;
}

static void
int_on_write_would_not_block(verto_ctx *ctx, verto_ev *ev)
{
  jsonrpc *rpc;
  ssize_t byteswritten, len;
  const char *tmp;

  rpc = verto_get_private(ev);
  if (!rpc || !(tmp = dataqueue_peek(rpc->io.queues.awaiting_write))) {
    verto_set_flags(ev, verto_get_flags(ev) & ~VERTO_EV_FLAG_IO_WRITE);
    return;
  }

  /* Attempt to write out our msg_state */
  len = strlen(tmp);
  byteswritten = write(rpc->io.out.fd, tmp, len);

  /* On full write, free this msg_state. */
  if (byteswritten == len) {
    free(dataqueue_pop(rpc->io.queues.awaiting_write));

  /* Socket error, close the socket */
  } else if (byteswritten < 0 && errno != EAGAIN && errno != EWOULDBLOCK) {
    rpc->errcb(rpc->io.in.ev, errno, rpc->misc);

  /* Partial write, save the rest for the next write opportunity */
  } else if (byteswritten < len)
    dataqueue_partial(rpc->io.queues.awaiting_write, byteswritten);
}

static int
int_write_data(jsonrpc *rpc, const char *data, bool newl)
{
  verto_ev_flag flags;

  if (!rpc)
    return EINVAL;

  flags = verto_get_flags(rpc->io.out.ev);
  if (!(flags & VERTO_EV_FLAG_IO_WRITE))
    verto_set_flags(rpc->io.out.ev, flags | VERTO_EV_FLAG_IO_WRITE);

  if (data && !dataqueue_push(rpc->io.queues.awaiting_write, newl, data))
    return ENOMEM;

  return 0;
}

static void
int_on_reader_free(verto_ctx *ctx, verto_ev *ev)
{
  int_jsonrpc_free(verto_get_private(ev));
}

static int
int_dispatch_rpcobject_request(jsonrpc *rpc, rpcobject *rpco,
                               rpcobject_type type, json_object **response)
{
  handler *dflt = NULL, *hndl = NULL;
  json_object *tmp = NULL;
  size_t i;

  assert(type & rpcobject_type_request);

  if (type & rpcobject_type_call) {
    json_object *id;

    *response = json_tokener_parse("{ \"jsonrpc\": \"2.0\" }");
    if (!*response)
      return ENOMEM;

    id = json_object_get(rpcobject_get_id(rpco));
    if (!id) {
      json_object_put(*response);
      return ENOMEM;
    }

    json_object_object_add(*response, "id", id);
  }

  /* Find the default and method-based handlers */
  for (i = 0; i < rpc->handlers.used; i++) {
    if (!rpc->handlers.data[i]->method) {
      if (!dflt)
        dflt = rpc->handlers.data[i];
    } else if (!strcmp(rpc->handlers.data[i]->method,
                       rpcobject_get_method(rpco))) {
      hndl = rpc->handlers.data[i];
      break;
    }
  }

  /* If no method match was found, call the default handler */
  if (!hndl)
    hndl = dflt;

  /* Call the handler */
  if (hndl) {
    if (hndl->handler(verto_get_ctx(rpc->io.in.ev), rpcobject_get_method(rpco),
                      rpcobject_get_params(rpco), hndl->misc,
                      (type & rpcobject_type_call) ? &tmp : NULL)) {
      if (type & rpcobject_type_call)
        json_object_object_add(*response, "result", tmp);
    } else {
      if (!tmp) {
        json_object_put(*response);
        return ENOMEM;
      }

      /* Set the error response */
      json_object_object_add(*response, "error", tmp);

      /* Validate that we have a valid error object */
      rpco = rpcobject_new(*response);
      if (!rpco) {
        json_object_put(*response);
        return ENOMEM;
      }
      if (!(rpcobject_get_type(rpco) & rpcobject_type_error)) {
        rpcobject_free(rpco);
        json_object_put(*response);
        return EINVAL;
      }
      rpcobject_free(rpco);
    }
  } else if (type & rpcobject_type_call) {
    tmp = int_get_error(NULL, NULL, JSONRPC_ERRCODE_METHOD_NOT_FOUND, true);
    if (!tmp) {
      json_object_put(*response);
      return ENOMEM;
    }
    json_object_object_add(*response, "error", tmp);
  }

  return 0;
}

static int
int_dispatch_rpcobject_response(jsonrpc *rpc, rpcobject *rpco,
                                rpcobject_type type, json_object **response)
{
  bool error;

  assert(type & rpcobject_type_response);
  error = type & rpcobject_type_error;

  if (rpcobject_get_id_type(rpco) == json_type_int)
    callqueue_respond(rpc->io.queues.awaiting_response,
                      rpcobject_get_id_int(rpco),
                      error ? rpcobject_get_error(rpco)
                            : rpcobject_get_result(rpco),
                      error);

  /* Silently drop responses with non-integer id's since they are errors that
   * we have no potential to fix */
  /* Silently drop responses without oustanding calls */

  return 0;
}

static int
int_dispatch_rpcobject(jsonrpc *rpc, rpcobject *rpco, json_object **response)
{
  rpcobject_type type = rpcobject_get_type(rpco);

  if (type & rpcobject_type_request)
    return int_dispatch_rpcobject_request(rpc, rpco, type, response);

  else if (type & rpcobject_type_response)
    return int_dispatch_rpcobject_response(rpc, rpco, type, response);

  *response = int_get_error(NULL, NULL, JSONRPC_ERRCODE_INVALID_REQUEST, false);
  return *response ? 0 : ENOMEM;
}

static int
int_dispatch_valid_json(jsonrpc *rpc, json_object *json)
{
  json_object *error = NULL;
  json_object *response = NULL;
  rpcobject *rpco = NULL;
  const char *errstr = JSONRPC_ERROR_INTERR;
  int retval = 0;

  assert(rpc && json);

  if (json_object_is_type(json, json_type_array)) {
    int i, len;

    response = json_object_new_array();
    if (!response) {
      retval = ENOMEM;
      goto internal_error;
    }

    len = json_object_array_length(json);
    for (i=0 ; i < len ; i++) {
      json_object *tmp = NULL;

      rpco = rpcobject_new(json_object_array_get_idx(json, i));
      if (!rpco) {
        retval = ENOMEM;
        json_object_put(response);
        goto internal_error;
      }

      retval = int_dispatch_rpcobject(rpc, rpco, &tmp);
      rpcobject_free(rpco);
      if (retval) {
        json_object_put(response);
        goto internal_error;
      }

      if (tmp) {
        retval = json_object_array_add(response, tmp);
        if (retval != 0) {
          retval = ENOMEM;
          json_object_put(response);
          goto internal_error;
        }
      }
    }

    len = json_object_array_length(response);
    if (len == 0) {
      json_object_put(response);
      response = NULL;
    }

    goto egress;

  } else if (json_object_is_type(json, json_type_object)) {
    rpco = rpcobject_new(json);
    if (!rpco) {
      retval = ENOMEM;
      goto internal_error;
    }

    retval = int_dispatch_rpcobject(rpc, rpco, &response);
    rpcobject_free(rpco);
    if (retval)
      goto internal_error;

    goto egress;
  }

  /* We have a non-array/non-object, so report the error */
  error = int_get_error(NULL, NULL, JSONRPC_ERRCODE_INVALID_REQUEST, false);
  if (!error || !(errstr = json_object_to_json_string(error))) {
    retval = ENOMEM;
    goto internal_error;
  }
  retval = int_write_data(rpc, errstr, true);
  json_object_put(error);
  if (!retval)
    return 0;

internal_error:
  int_write_data(rpc, JSONRPC_ERROR_INTERR, false);
  rpc->errcb(rpc->io.in.ev, retval, rpc->misc);
  return retval;

egress:
  if (response) {
    retval = int_write_data(rpc, json_object_to_json_string(response), true);
    json_object_put(response);
    if (retval)
      goto internal_error;
  }

  return 0;
}

static void
int_on_read_would_not_block(verto_ctx *ctx, verto_ev *ev)
{
  ssize_t bytesread;
  jsonrpc *rpc;
  char *msgterm;

  rpc = verto_get_private(ev);
  if (!rpc)
    return;

  /* Read into our buffer */
  bytesread = read(rpc->io.in.fd, &rpc->io.buffer.data[rpc->io.buffer.used],
                   rpc->io.buffer.size - rpc->io.buffer.used);
  if (bytesread < 1) {
    rpc->errcb(rpc->io.in.ev, errno, rpc->misc);
    return;
  }
  rpc->io.buffer.used += bytesread;
  rpc->io.buffer.data[rpc->io.buffer.used] = '\0';

  /* Parse objects */
  while ((msgterm = strchr(rpc->io.buffer.data, JSONRPC_MESSAGE_TERMINATOR))) {
    json_object *json = NULL;

    /* Parse the msg_state */
    json_tokener_reset(rpc->parser.tokener);
    json = json_tokener_parse_ex(rpc->parser.tokener, rpc->io.buffer.data,
                                 msgterm - rpc->io.buffer.data);

    /* Shift the buffer */
    rpc->io.buffer.used -= msgterm - rpc->io.buffer.data + 1;
    memmove(rpc->io.buffer.data, msgterm + 1, rpc->io.buffer.used);
    rpc->io.buffer.data[rpc->io.buffer.used] = '\0';

    /* Skip one request if msg is too large */
    if (rpc->parser.msgtoolarge)
      rpc->parser.msgtoolarge = false;

    /* Otherwise attempt to dispatch the message */
    else {
      if (json) {
        if (int_dispatch_valid_json(rpc, json) != 0) {
          json_object_put(json);
          break;
        }
      } else /* If we failed to parse the message, fire a PARSE ERROR */
        int_write_data(rpc, JSONRPC_ERROR_PARSEERR, false);
    }

    json_object_put(json);
  }

  /* If we can't find the input terminator and the buffer is full then
   * we are over our max input size. Setup so that further reads will
   * discard data until the next input terminator + 1 and fire an error. */
  if (rpc->io.buffer.used >= rpc->io.buffer.size) {
    rpc->parser.msgtoolarge = true;
    rpc->io.buffer.used = 0;
    int_write_data(rpc, JSONRPC_ERROR_TOOBIG, false);
  }
}

static void
int_on_io_would_not_block(verto_ctx *ctx, verto_ev *ev)
{
  verto_ev_flag state = verto_get_fd_state(ev);

  if (state & VERTO_EV_FLAG_IO_READ)
    int_on_read_would_not_block(ctx, ev);
  if (state & VERTO_EV_FLAG_IO_WRITE)
    int_on_write_would_not_block(ctx, ev);
}

verto_ev *
jsonrpc_init(verto_ctx *ctx, int infd, int outfd, size_t maxsize,
             jsonrpc_error_cb *errcb, void *misc, jsonrpc_free *free)
{
  jsonrpc *rpc;

  if (!ctx || !errcb)
    return NULL;

  rpc = malloc(sizeof(jsonrpc));
  if (!rpc)
    return NULL;
  memset(rpc, 0, sizeof(jsonrpc));
  rpc->errcb = errcb;
  rpc->misc  = misc;
  rpc->free  = free;
  rpc->io.in.fd = infd;
  rpc->io.in.ev = verto_add_io(ctx, VERTO_EV_FLAG_IO_READ
                                      | VERTO_EV_FLAG_IO_ERROR
                                      | VERTO_EV_FLAG_PERSIST
                                      | VERTO_EV_FLAG_IO_CLOSE_FD,
                               int_on_io_would_not_block, infd);
  rpc->io.out.fd = outfd;
  rpc->io.buffer.size = maxsize;
  rpc->io.buffer.data = malloc(maxsize);
  rpc->io.queues.awaiting_write = dataqueue_new();
  rpc->io.queues.awaiting_response = callqueue_new(rpc->io.queues.awaiting_write);
  rpc->parser.tokener = json_tokener_new();

  if (infd == outfd)
    rpc->io.out.ev = rpc->io.in.ev;
  else
    rpc->io.out.ev = verto_add_io(ctx, VERTO_EV_FLAG_IO_WRITE
                                         | VERTO_EV_FLAG_IO_ERROR
                                         | VERTO_EV_FLAG_PERSIST
                                         | VERTO_EV_FLAG_IO_CLOSE_FD,
                                  int_on_io_would_not_block, outfd);

  if (!rpc->io.in.ev || !rpc->io.out.ev || !rpc->io.buffer.data
      || !rpc->io.queues.awaiting_write || !rpc->io.queues.awaiting_response
      || !rpc->parser.tokener) {
    int_jsonrpc_free(rpc);
    return NULL;
  }

  verto_set_private(rpc->io.in.ev, rpc, int_on_reader_free);
  if (rpc->io.in.ev != rpc->io.out.ev)
    verto_set_private(rpc->io.out.ev, rpc, int_on_writer_free);
  return rpc->io.in.ev;
}

bool
jsonrpc_set_handler(verto_ev *ev, const char *method,
    jsonrpc_handler_cb *handler_cb, void *misc, jsonrpc_free *mfree)
{
  handler *hentry;
  jsonrpc *rpc;

  if (!ev || !(rpc = verto_get_private(ev)))
    return false;

  if (!handler_cb) {
    /* We want to delete the handler */
    for (size_t i = 0; i < rpc->handlers.used; i++) {
      if (method == rpc->handlers.data[i]->method
          || (method && rpc->handlers.data[i]->method
              && !strcmp(method, rpc->handlers.data[i]->method))) {

        /* Free the existing handler */
        int_handler_free(rpc->handlers.data[i]);

        /* Shift down the other handlers */
        memmove(&rpc->handlers.data[i], &rpc->handlers.data[i + 1],
            rpc->handlers.used-- - i);

        /* Shrink if we need to */
        if (rpc->handlers.used > 0
            && rpc->handlers.used < rpc->handlers.size / 4) {
          handler **tmp;

          tmp = realloc(rpc->handlers.data,
              sizeof(void*) * rpc->handlers.size / 2);
          if (tmp) {
            rpc->handlers.size /= 2;
            rpc->handlers.data = tmp;
          }
        }

        return true;
      }
    }
  }

  /* Create our new entry */
  hentry = malloc(sizeof(handler));
  if (!hentry)
    return false;
  hentry->misc = misc;
  hentry->free = mfree;
  hentry->handler = handler_cb;
  hentry->method = method ? strdup(method) : NULL;
  if (method && !hentry->method)
    goto error;

  /* Replace a matching method */
  for (size_t i = 0; i < rpc->handlers.used; i++) {
    if (method == rpc->handlers.data[i]->method
        || (method && rpc->handlers.data[i]->method
            && !strcmp(method, rpc->handlers.data[i]->method))) {
      int_handler_free(rpc->handlers.data[i]);
      rpc->handlers.data[i] = hentry;
      return true;
    }
  }

  /* Grow the array if we need */
  if (rpc->handlers.used == rpc->handlers.size) {
    handler **tmp;

    rpc->handlers.size = (rpc->handlers.size > 4 ? rpc->handlers.size : 4);
    tmp = realloc(rpc->handlers.data, sizeof(void*) * rpc->handlers.size * 2);
    if (!tmp) {
      rpc->handlers.size = rpc->handlers.used;
      goto error;
    }

    rpc->handlers.data = tmp;
  }

  /* Insert our new handler */
  rpc->handlers.data[rpc->handlers.used++] = hentry;
  return true;

error:
  int_handler_free(hentry);
  return false;
}

jsonrpc_batch *
jsonrpc_batch_new(verto_ev *ev)
{
  jsonrpc_batch *batch = NULL;

  batch = malloc(sizeof(jsonrpc_batch));
  if (batch) {
    memset(batch, 0, sizeof(jsonrpc_batch));

    batch->rpc = verto_get_private(ev);
    if (batch->rpc) {
      batch->cq = callqueue_new(NULL);
      if (batch->cq) {
        batch->next = batch->rpc->batches;
        batch->rpc->batches = batch;
        return batch;
      }
    }
  }

  free(batch);
  return NULL;
}

bool
jsonrpc_batch_send(jsonrpc_batch *batch)
{
  if (int_write_data(batch->rpc, callqueue_as_batch(batch->cq), true) != 0)
    return false;

  callqueue_remove_notifications(batch->cq);
  callqueue_move(batch->cq, batch->rpc->io.queues.awaiting_response);
  return true;
}

void
jsonrpc_batch_free(jsonrpc_batch *batch)
{
  jsonrpc_batch *tmp, *prev = NULL;

  if (batch) {
    callqueue_free(batch->cq);
    for (tmp = batch->rpc->batches; tmp; prev = tmp, tmp = tmp->next) {
      if (tmp == batch) {
        if (prev)
          prev = batch->next;
        else
          batch->rpc->batches = batch->next;
        break;
      }
    }
    free(batch);
  }
}

bool
jsonrpc_call(verto_ev *ev, jsonrpc_batch *batch, const char *method,
    json_object *params, jsonrpc_response_cb *resp, void *misc,
    jsonrpc_free *mfree)
{
  jsonrpc *rpc;

  if (!(rpc = verto_get_private(ev))
      || !callqueue_push(batch ? batch->cq : rpc->io.queues.awaiting_response,
                         rpc->serial++, method, params, resp, misc, mfree)) {
    if (params)
      json_object_put(params);
    return false;
  }

  int_write_data(rpc, NULL, false);
  return true;
}
