/*
 * Copyright 2011 Red Hat, Inc.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include "rpcobject.h"
#include <stdlib.h>
#include <string.h>

struct rpcobject {
  rpcobject_type type;
  int id;
  json_object *idobj;

  union {
    struct {
      json_object *params;
      char *method;
    } request;

    union {
      json_object *result;
      struct {
        json_object *object;
        int code;
        char *message;
        json_object *data;
      } error;
    } response;
  } x;
};

static inline char *
xstrdup(const char *str)
{
  if (!str)
    return NULL;
  return strdup(str);
}

rpcobject *
rpcobject_new(json_object *obj)
{
  json_object *jsonrpc, *method, *params, *result, *error, *id;
  json_object *code = NULL, *message = NULL;
  const char *tmp;
  rpcobject *rpco;

  rpco = malloc(sizeof(rpcobject));
  if (!rpco)
    goto internal_error;
  memset(rpco, 0, sizeof(rpcobject));
  rpco->type = rpcobject_type_unknown;

  if (!obj)
    return rpco;

  jsonrpc = json_object_object_get(obj, "jsonrpc");
  method = json_object_object_get(obj, "method");
  params = json_object_object_get(obj, "params");
  result = json_object_object_get(obj, "result");
  error = json_object_object_get(obj, "error");
  id = json_object_object_get(obj, "id");
  if (error) {
    code = json_object_object_get(error, "code");
    message = json_object_object_get(error, "message");
  }

  /* Make sure that jsonrpc exists and is a string */
  if (!jsonrpc || !json_object_is_type(jsonrpc, json_type_string))
    return rpco;

  /* Make sure it has the proper version */
  if (!(tmp = json_object_get_string(jsonrpc)) || strcmp(tmp, JSONRPC_VERSION))
    return rpco;

  /* If id exists, it must be a number, a string or null. */
  if (id && !json_object_is_type(id, json_type_double)
      && !json_object_is_type(id, json_type_int)
      && !json_object_is_type(id, json_type_null)
      && !json_object_is_type(id, json_type_string))
    return rpco;

  /* We have a request */
  if (method && json_object_is_type(method, json_type_string)
             && (!params || json_object_is_type(params, json_type_array)
                         || json_object_is_type(params, json_type_object))) {
    /* Set the type */
    rpco->type = rpcobject_type_request;
    if (id)
      rpco->type |= rpcobject_type_call;
    else
      rpco->type |= rpcobject_type_notification;

    /* We probably have a response */
  } else {
    /* Make sure that only result or error exists, not both. */
    if ((result && error) || (!result && !error))
      return rpco;

    /* If this is an error response, verify that we have a valid error object. */
    if (error) {
      /* Make sure we have code and message. */
      if (!code || !message)
        return rpco;

      /* Make sure the code is an int */
      if (!json_object_is_type(code, json_type_int))
        return rpco;

      /* Make sure the message is a string */
      if (!json_object_is_type(message, json_type_string))
        return rpco;
    }

    /* Set the type */
    rpco->type = rpcobject_type_response;
    if (error)
      rpco->type |= rpcobject_type_error;
    else
      rpco->type |= rpcobject_type_result;
  }

  if (id) {
    rpco->idobj = json_object_get(id);
    if (json_object_is_type(id, json_type_int))
      rpco->id = json_object_get_int(id);
  }

  if (rpco->type & rpcobject_type_request) {
    rpco->x.request.method = xstrdup(json_object_get_string(method));
    if (!rpco->x.request.method)
      goto internal_error;
    rpco->x.request.params = json_object_get(params);
  } else if (rpco->type & rpcobject_type_response) {
    if (error) {
      rpco->x.response.error.message = xstrdup(json_object_get_string(message));
      if (!rpco->x.response.error.message)
        goto internal_error;
      rpco->x.response.error.object = json_object_get(error);
      rpco->x.response.error.code = json_object_get_int(code);
      rpco->x.response.error.data = json_object_object_get(error, "data");
      if (rpco->x.response.error.data)
        json_object_get(rpco->x.response.error.data);
    } else
      rpco->x.response.result = json_object_get(result);
  }

  return rpco;

internal_error:
  free(rpco);
  return NULL;
}

rpcobject_type
rpcobject_get_type(rpcobject *obj)
{
  return obj ? obj->type : rpcobject_type_unknown;
}

bool
rpcobject_is_type(rpcobject *obj, rpcobject_type types)
{
  return (obj ? obj->type : rpcobject_type_unknown) & types;
}

json_object *
rpcobject_get_id(rpcobject *obj)
{
  return obj ? obj->idobj : NULL;
}

int
rpcobject_get_id_int(rpcobject *obj)
{
  return obj ? obj->id : 0;
}

json_type
rpcobject_get_id_type(rpcobject *obj)
{
  return obj && obj->idobj ? json_object_get_type(obj->idobj) : json_type_null;
}

const char *
rpcobject_get_method(rpcobject *obj)
{
  if (!obj || !(obj->type & rpcobject_type_request))
    return NULL;
  return obj->x.request.method;
}

json_object *
rpcobject_get_params(rpcobject *obj)
{
  if (!obj || !(obj->type & rpcobject_type_request))
    return NULL;
  return obj->x.request.params;
}

json_object *
rpcobject_get_result(rpcobject *obj)
{
  if (!obj || !(obj->type & rpcobject_type_result))
    return NULL;
  return obj->x.response.result;
}

json_object *
rpcobject_get_error(rpcobject *obj)
{
  if (!obj || !(obj->type & rpcobject_type_error))
    return NULL;
  return obj->x.response.error.object;
}

int
rpcobject_get_error_code(rpcobject *obj)
{
  if (!obj || !(obj->type & rpcobject_type_error))
    return 0;
  return obj->x.response.error.code;
}

const char *
rpcobject_get_error_message(rpcobject *obj)
{
  if (!obj || !(obj->type & rpcobject_type_error))
    return NULL;
  return obj->x.response.error.message;
}

json_object *
rpcobject_get_error_data(rpcobject *obj)
{
  if (!obj || !(obj->type & rpcobject_type_error))
    return NULL;
  return obj->x.response.error.data;
}

void
rpcobject_free(rpcobject *obj)
{
  if (!obj)
    return;

  json_object_put(obj->idobj);
  if (obj->type & rpcobject_type_request) {
    free(obj->x.request.method);
    json_object_put(obj->x.request.params);
  } else if (obj->type & rpcobject_type_error) {
    free(obj->x.response.error.message);
    json_object_put(obj->x.response.error.data);
    json_object_put(obj->x.response.error.object);
  } else if (obj->type & rpcobject_type_result)
    json_object_put(obj->x.response.result);
  free(obj);
}
